@@ -1,5 +1,5 @@
 // SPDX-License-Identifier: LGPL-2.1-only
-// Copyright (c) 2015 - 2019 DisplayLink (UK) Ltd.
+// Copyright (c) 2015 - 2020 DisplayLink (UK) Ltd.
 
 #include <stddef.h>
 #include <stdint.h>
@@ -29,7 +29,7 @@
 
 #define EVDI_MODULE_COMPATIBILITY_VERSION_MAJOR 1
 #define EVDI_MODULE_COMPATIBILITY_VERSION_MINOR 5
-#define EVDI_MODULE_COMPATIBILITY_VERSION_PATCHLEVEL 0
+#define EVDI_MODULE_COMPATIBILITY_VERSION_PATCH 0
 
 #define evdi_log(...) do {						\
 	if (g_evdi_logging.function) {					\
@@ -83,18 +83,18 @@ static int drm_auth_magic(int fd, drm_magic_t magic)
 static int drm_is_master(int fd)
 {
 	/* Detect master by attempting something that requires master.
-	*
-	* Authenticating magic tokens requires master and 0 is an
-	* internal kernel detail which we could use. Attempting this on
-	* a master fd would fail therefore fail with EINVAL because 0
-	* is invalid.
-	*
-	* A non-master fd will fail with EACCES, as the kernel checks
-	* for master before attempting to do anything else.
-	*
-	* Since we don't want to leak implementation details, use
-	* EACCES.
-	*/
+	 *
+	 * Authenticating magic tokens requires master and 0 is an
+	 * internal kernel detail which we could use. Attempting this on
+	 * a master fd would fail therefore fail with EINVAL because 0
+	 * is invalid.
+	 *
+	 * A non-master fd will fail with EACCES, as the kernel checks
+	 * for master before attempting to do anything else.
+	 *
+	 * Since we don't want to leak implementation details, use
+	 * EACCES.
+	 */
 	return drm_auth_magic(fd, 0) != -EACCES;
 }
 
@@ -181,7 +181,7 @@ static int is_evdi_compatible(int fd)
 	evdi_log("LibEvdi version (%d.%d.%d)",
 		 LIBEVDI_VERSION_MAJOR,
 		 LIBEVDI_VERSION_MINOR,
-		 LIBEVDI_VERSION_PATCHLEVEL);
+		 LIBEVDI_VERSION_PATCH);
 
 	if (do_ioctl(fd, DRM_IOCTL_VERSION, &ver, "version") != 0)
 		return 0;
@@ -198,7 +198,7 @@ static int is_evdi_compatible(int fd)
 	evdi_log("Doesn't match LibEvdi compatibility one (%d.%d.%d)",
 		 EVDI_MODULE_COMPATIBILITY_VERSION_MAJOR,
 		 EVDI_MODULE_COMPATIBILITY_VERSION_MINOR,
-		 EVDI_MODULE_COMPATIBILITY_VERSION_PATCHLEVEL);
+		 EVDI_MODULE_COMPATIBILITY_VERSION_PATCH);
 
 	return 0;
 }
@@ -357,24 +357,27 @@ static int open_as_slave(const char *device_path)
 {
 	int fd = 0;
 	int err = 0;
+
 	fd = open(device_path, O_RDWR);
-	if (fd < 0) {
+	if (fd < 0)
 		return -1;
-	}
 
 	if (drm_is_master(fd)) {
-		evdi_log("Process has master on %s, err: %s", device_path, strerror(errno));
+		evdi_log("Process has master on %s, err: %s",
+			 device_path, strerror(errno));
 		err = drm_ioctl(fd, DRM_IOCTL_DROP_MASTER, NULL);
 	}
 
 	if (err < 0) {
-		evdi_log("Drop master on %s failed, err: %s", device_path, strerror(errno));
+		evdi_log("Drop master on %s failed, err: %s",
+			 device_path, strerror(errno));
 		close(fd);
 		return err;
 	}
 
 	if (drm_is_master(fd)) {
-		evdi_log("Drop master on %s failed, err: %s", device_path, strerror(errno));
+		evdi_log("Drop master on %s failed, err: %s",
+			 device_path, strerror(errno));
 		close(fd);
 		return -1;
 	}
@@ -445,12 +448,12 @@ evdi_handle evdi_open(int device)
 	return h;
 }
 
-enum evdi_device_status evdi_check_device(int device)
+static enum evdi_device_status evdi_device_to_platform(int device, char *path)
 {
 	struct dirent *fd_entry;
 	DIR *fd_dir;
 	enum evdi_device_status status = UNRECOGNIZED;
-	char path[PATH_MAX];
+	char card_path[PATH_MAX];
 
 	if (!device_exists(device))
 		return NOT_PRESENT;
@@ -466,10 +469,9 @@ enum evdi_device_status evdi_check_device(int device)
 			continue;
 
 		snprintf(path, PATH_MAX,
-			"/sys/devices/platform/%s/drm/card%d",
-			fd_entry->d_name,
-			device);
-		if (path_exists(path)) {
+			"/sys/devices/platform/%s", fd_entry->d_name);
+		snprintf(card_path, PATH_MAX, "%s/drm/card%d", path, device);
+		if (path_exists(card_path)) {
 			status = AVAILABLE;
 			break;
 		}
@@ -479,6 +481,13 @@ enum evdi_device_status evdi_check_device(int device)
 	return status;
 }
 
+enum evdi_device_status evdi_check_device(int device)
+{
+	char path[PATH_MAX];
+
+	return evdi_device_to_platform(device, path);
+}
+
 int evdi_add_device(void)
 {
 	FILE *add_devices = fopen("/sys/devices/evdi/add", "w");
@@ -529,6 +538,41 @@ void evdi_disconnect(evdi_handle handle)
 	do_ioctl(handle->fd, DRM_IOCTL_EVDI_CONNECT, &cmd, "disconnect");
 }
 
+void evdi_enable_cursor_events(evdi_handle handle)
+{
+	char path[PATH_MAX] = {0};
+	static const char enable[] = "Y";
+	int path_len = 0;
+	FILE *cursor_evs = NULL;
+	size_t written = 0;
+	const size_t elem_bytes = 1;
+	int errcode = 0;
+
+	if (evdi_device_to_platform(handle->device_index, path) !=
+	    AVAILABLE) {
+		evdi_log("Failed to enable cursor events");
+		evdi_log("Device /dev/dri/card%d, device is not available.",
+			handle->device_index);
+		return;
+	}
+
+	path_len = strlen(path);
+	snprintf(path+path_len, PATH_MAX-path_len, "/cursor_events");
+	cursor_evs = fopen(path, "w");
+	if (cursor_evs == NULL) {
+		evdi_log("Failed to open %s, err: %s", path, strerror(errno));
+		return;
+	}
+
+	written = fwrite(enable, elem_bytes, sizeof(enable), cursor_evs);
+	errcode = errno;
+	fclose(cursor_evs);
+	evdi_log("Enabling cursor events on /dev/dri/card%d %s %s",
+		handle->device_index,
+		written < sizeof(enable) ? "failed: " : "succeeded",
+		written < sizeof(enable) ? strerror(errcode) : "");
+}
+
 void evdi_grab_pixels(evdi_handle handle,
 		      struct evdi_rect *rects,
 		      int *num_rects)
@@ -785,7 +829,7 @@ void evdi_get_lib_version(struct evdi_lib_version *version)
 	if (version != NULL) {
 		version->version_major = LIBEVDI_VERSION_MAJOR;
 		version->version_minor = LIBEVDI_VERSION_MINOR;
-		version->version_patchlevel = LIBEVDI_VERSION_PATCHLEVEL;
+		version->version_patchlevel = LIBEVDI_VERSION_PATCH;
 	}
 }
 
