@@ -1,7 +1,7 @@
 // SPDX-License-Identifier: GPL-2.0-only
 /*
  * Copyright (C) 2012 Red Hat
- * Copyright (c) 2015 - 2019 DisplayLink (UK) Ltd.
+ * Copyright (c) 2015 - 2020 DisplayLink (UK) Ltd.
  *
  * Based on parts on udlfb.c:
  * Copyright (C) 2009 its respective authors
@@ -11,12 +11,11 @@
  * more details.
  */
 
-#include <drm/drmP.h>
+#include <linux/version.h>
 #include <drm/drm_crtc.h>
 #include <drm/drm_edid.h>
 #include <drm/drm_crtc_helper.h>
 #include <drm/drm_atomic_helper.h>
-#include <linux/version.h>
 #include "evdi_drv.h"
 
 #if KERNEL_VERSION(5, 1, 0) <= LINUX_VERSION_CODE
@@ -52,7 +51,7 @@ static int evdi_get_modes(struct drm_connector *connector)
 #endif
 
 	if (!ret)
-		drm_add_edid_modes(connector, edid);
+		ret = drm_add_edid_modes(connector, edid);
 	else
 		EVDI_ERROR("Failed to set edid modes! error: %d", ret);
 
@@ -106,12 +105,19 @@ static void evdi_connector_destroy(struct drm_connector *connector)
 
 static struct drm_encoder *evdi_best_encoder(struct drm_connector *connector)
 {
+#if KERNEL_VERSION(5, 5, 0) <= LINUX_VERSION_CODE
+	struct drm_encoder *encoder;
 
+	drm_connector_for_each_possible_encoder(connector, encoder) {
+		return encoder;
+	}
+
+	return NULL;
+#else
 	return drm_encoder_find(connector->dev,
-#if KERNEL_VERSION(4, 15, 0) <= LINUX_VERSION_CODE
 				NULL,
-#endif
 				connector->encoder_ids[0]);
+#endif
 }
 
 static struct drm_connector_helper_funcs evdi_connector_helper_funcs = {
@@ -121,15 +127,9 @@ static struct drm_connector_helper_funcs evdi_connector_helper_funcs = {
 };
 
 static const struct drm_connector_funcs evdi_connector_funcs = {
-#if KERNEL_VERSION(4, 14, 0) > LINUX_VERSION_CODE
-	.dpms = drm_atomic_helper_connector_dpms,
-#endif
 	.detect = evdi_detect,
 	.fill_modes = drm_helper_probe_single_connector_modes,
 	.destroy = evdi_connector_destroy,
-#if KERNEL_VERSION(4, 14, 0) > LINUX_VERSION_CODE
-	.set_property = drm_atomic_helper_connector_set_property,
-#endif
 	.reset = drm_atomic_helper_connector_reset,
 	.atomic_duplicate_state = drm_atomic_helper_connector_duplicate_state,
 	.atomic_destroy_state = drm_atomic_helper_connector_destroy_state
@@ -156,11 +156,5 @@ int evdi_connector_init(struct drm_device *dev, struct drm_encoder *encoder)
 #else
 	drm_mode_connector_attach_encoder(connector, encoder);
 #endif
-
-#if KERNEL_VERSION(4, 9, 0) > LINUX_VERSION_CODE
-	drm_object_attach_property(&connector->base,
-				   dev->mode_config.dirty_info_property, 1);
-#endif
-
 	return 0;
 }
